# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from os import path
from random import choice, sample


def get_pos(from_pos, direction, columns, rows):
    if direction == "N":
        if from_pos - columns < 0:
            return -1
        return from_pos - columns
    if direction == "S":
        if from_pos + columns >= columns * rows:
            return -1
        return from_pos + columns
    if direction == "E":
        if from_pos // columns != (from_pos + 1) // columns:
            return -1
        return from_pos + 1
    if direction == "W":
        if from_pos // columns != (from_pos - 1) // columns:
            return -1
        return from_pos - 1


def min_distance(pos, food, config):
    cols = config.columns
    return min([abs(pos % cols - fpos % cols) + abs(pos // cols - fpos // cols) for fpos in food])


def random_agent():
    return choice(["N", "S", "E", "W"])


def shortest_path_agent(obs, config):
    columns = config.columns
    rows = config.rows
    goose = obs.geese[obs.index]
    head = goose[0]
    max_value = columns * rows
    directions = ["N", "S", "E", "W"]

    geese = [g for g in obs.geese if len(g) > 0]
    heads = [g[0] for g in geese]
    tails = [g[-1] for g in geese]
    bodies = [p for g in geese for p in g[:-1]]

    actions = {}
    for d in directions:
        # Get new position - or -1 if run into wall.
        pos = get_pos(head, d, columns, rows)
        if (
            pos in bodies or  # Hit a body.
            (len(goose) > 1 and goose[1] == pos) or  # Backwards.
            # Hit a tail when head over food.
            (pos in tails and heads[tails.index(pos)] in obs.food)
        ):
            pos = -1
        actions[d] = max_value if pos == - \
            1 else min_distance(pos, obs.food, config)
        # Possibility of a collision, devalue a valid action.
        for h in heads:
            if h == head:
                continue
            for dh in directions:
                posh = get_pos(h, dh, columns, rows)
                if posh == pos:
                    actions[d] += 1

    return min(actions, key=actions.get)


agents = {"random": random_agent, "shortest": shortest_path_agent}


def interpreter(state, env):
    config = env.configuration
    columns = config.columns
    rows = config.rows
    hunger_rate = config.hunger_rate
    min_food = config.min_food
    num_agents = len(state)

    # Clone the geese and food observation between all the agents.
    geese = state[0].observation.geese
    food = state[0].observation.food
    for agent in state:
        agent.observation.geese = geese
        agent.observation.food = food

    # Reset the environment.
    if env.done:
        # Distribute food and geese randomly.
        starting_positions = sample(range(columns * rows), num_agents * 2)
        for index in range(num_agents):
            geese.append([starting_positions[index]])
            food.append(starting_positions[index + num_agents])
        return state

    # Update active agents rewards.
    for index, agent in enumerate(state):
        if agent.status == "ACTIVE":
            agent.reward = len(env.steps) + len(geese[index])

    # Apply the actions from active agents.
    for index, agent in enumerate(state):
        if agent.status != "ACTIVE":
            continue
        action = agent.action
        goose = geese[index]
        head = goose[0]
        new_head = get_pos(head, action, columns, rows)

        # Wall Hit.
        if new_head == -1:
            env.debug_print(f"Wall Hit: {action}")
            agent.status = "INACTIVE"
            geese[index] = []
            continue

        # Last Body Hit.
        if len(goose) > 1 and goose[1] == new_head:
            env.debug_print(f"Body Hit: {action}")
            agent.status = "INACTIVE"
            geese[index] = []
            continue

        # Add New Head to the Goose.
        goose.insert(0, new_head)

        # Check Food.
        if head in food:
            food.remove(head)
        else:
            goose.pop()

        # If hunger strikes remove from the tail.
        if len(env.steps) % hunger_rate == 0:
            goose.pop()
            if len(goose) == 0:
                env.debug_print(f"Goose Starved: {action}")
                agent.status = "INACTIVE"
                geese[index] = []
                continue

    # Check for collisions.
    collisions = {}
    for goose in geese:
        for pos in goose:
            collisions[pos] = collisions.get(pos, 0) + 1
    for index, agent in enumerate(state):
        for pos in geese[index]:
            if collisions[pos] > 1:
                env.debug_print(f"Goose Collision: {agent.action}")
                agent.status = "INACTIVE"
                geese[index] = []
                continue

    # Add food if min_food threshold reached.
    if len(food) < min_food:
        available_positions = list(range(rows * columns))
        for goose in geese:
            for pos in goose:
                available_positions.remove(pos)
        food.extend(sample(available_positions, min_food - len(food)))

    # If only one ACTIVE agent left, set it to INACTIVE.
    active_agents = [a for a in state if a.status == "ACTIVE"]
    if len(active_agents) == 1:
        active_agents[0].status = "INACTIVE"

    return state


def renderer(state, env):
    config = env.configuration
    columns = config.columns
    rows = config.rows

    foodSymbol = "F"
    colDivider = "|"
    rowDivider = "+" + "+".join(["---"] * columns) + "+\n"

    board = [" "] * (rows * columns)
    for pos in state[0].observation.food:
        board[pos] = foodSymbol

    for index, goose in enumerate(state[0].observation.geese):
        for pos in goose:
            board[pos] = index

    out = rowDivider
    for row in range(rows):
        for col in range(columns):
            out += colDivider + f" {board[(row * columns) + col]} "
        out += colDivider + "\n" + rowDivider

    return out


dirpath = path.dirname(__file__)
jsonpath = path.abspath(path.join(dirpath, "battlegeese.json"))
with open(jsonpath) as f:
    specification = json.load(f)


def html_renderer():
    jspath = path.abspath(path.join(dirpath, "battlegeese.js"))
    with open(jspath) as f:
        return f.read()



from ..halite.helpers import ReadOnlyDict


class Observation(ReadOnlyDict[str, any]):
    """
    Observation primarily used as a helper to construct the State from the raw observation.
    This provides bindings for the observation type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/rps/rps.json
    """
    @property
    def last_opponent_action(self) -> int:
        """Move the opponent took on the last turn."""
        return self["halite"]

    @property
    def step(self) -> int:
        """The current step index within the episode."""
        return self["step"]


class Configuration(ReadOnlyDict[str, any]):
    """
    Configuration provides access to tunable parameters in the environment.
    This provides bindings for the configuration type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/rps/rps.json
    """
    @property
    def episode_steps(self) -> int:
        """Total number of steps in the episode."""
        return self["episodeSteps"]

    @property
    def signs(self) -> int:
        """Maximum runtime (seconds) to initialize an agent."""
        return self["agentTimeout"]

    @property
    def act_timeout(self) -> float:
        """Maximum runtime (seconds) to obtain an action from an agent."""
        return self["actTimeout"]

    @property
    def run_timeout(self) -> float:
        """Maximum runtime (seconds) of an episode (not necessarily DONE)."""
        return self["runTimeout"]



# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import jsonschema
from copy import deepcopy
from pathlib import Path
from .errors import InvalidArgument, NotFound


# Path Utilities.
root_path = Path(__file__).parent.resolve()
envs_path = Path.joinpath(root_path, "envs")


# Primitive Utilities.
def get(o, classinfo=None, default=None, path=[], is_callable=None, fallback=None):
    if o is None and default is not None:
        o = default
    if has(o, classinfo, default, path, is_callable):
        cur = o
        for p in path:
            cur = cur[p]
        return cur
    else:
        if default is not None:
            return default
        return fallback


def has(o, classinfo=None, default=None, path=[], is_callable=None):
    try:
        cur = o
        for p in path:
            cur = cur[p]
        if classinfo is not None and not isinstance(cur, classinfo):
            raise "Not a match"
        if is_callable == True and not callable(cur):
            raise "Not callable"
        if is_callable == False and callable(cur):
            raise "Is callable"
        return True
    except:
        if default is not None and o is not None and len(path) > 0:
            cur = o
            for p in path[:-1]:
                if not has(cur, dict, path=[p]):
                    cur[p] = {}
                cur = cur[p]
            cur[path[-1]] = default
        return False


def call(o, default=None, path=[], args=[], kwargs={}):
    o = get(o, default=False, path=path, is_callable=True)
    if o is not False:
        return o(*args, **kwargs)
    else:
        return default


class Struct(dict):
    def __init__(self, **entries):
        entries = {k: v for k, v in entries.items() if k != "items"}
        dict.__init__(self, entries)
        self.__dict__.update(entries)

    def __setattr__(self, attr, value):
        self.__dict__[attr] = value
        self[attr] = value


# Added benefit of cloning lists and dicts.
def structify(o):
    if isinstance(o, list):
        return [structify(o[i]) for i in range(len(o))]
    elif isinstance(o, dict):
        return Struct(**{k: structify(v) for k, v in o.items()})
    return o


# File Utilities.
def read_file(path, fallback=None):
    try:
        with open(path, "r", encoding="utf-8") as file:
            return file.read()
    except:
        if fallback is not None:
            return fallback
        raise NotFound(f"{path} not found")


def get_file_json(path, fallback=None):
    try:
        with open(path, "r") as json_file:
            return json.load(json_file)
    except:
        if fallback is not None:
            return fallback
        raise InvalidArgument(f"{path} does not contain valid JSON")


# Schema Utilities.
schemas = structify(get_file_json(Path.joinpath(root_path, "schemas.json")))


def default_schema(schema, data):
    default = get(schema, path=["default"])
    if default is None and data is None:
        return

    if get(schema, path=["type"]) == "object":
        default = deepcopy(get(default, dict, {}))
        if data is None or not has(data, dict):
            obj = default
        else:
            obj = data
            for k, v in default.items():
                if not k in obj:
                    obj[k] = v
        properties = get(schema, dict, {}, ["properties"])
        for key, prop_schema in properties.items():
            new_value = default_schema(prop_schema, get(obj, path=[key]))
            if new_value is not None:
                obj[key] = new_value
        return obj

    if get(schema, path=["type"]) == "array":
        default = deepcopy(get(default, list, []))
        arr = get(data, list, default)
        item_schema = get(schema, dict, {}, ["items"])
        for index, value in enumerate(arr):
            if value is None and len(default) > index:
                new_value = default[index]
            else:
                new_value = default_schema(item_schema, value)
            if new_value is not None:
                arr[index] = new_value
        return arr

    return data if data is not None else default


def process_schema(schema, data, use_default=True):
    error = None
    if use_default is True:
        data = default_schema(schema, deepcopy(data))
    try:
        jsonschema.validate(data, schema)
    except Exception as err:
        error = str(err)
    return error, data


# Player utilities
def get_player(window_kaggle, renderer):
    key = "/*window.kaggle*/"
    value = f"""
window.kaggle = {json.dumps(window_kaggle, indent=2)};\n\n
window.kaggle.renderer = {renderer.strip()};\n\n
    """
    return read_file(Path.joinpath(root_path, "static", "player.html")).replace(
        key, value
    )



from kaggle_environments import make, evaluate
from kaggle_environments.errors import DeadlineExceeded
import copy
import os

env = None


def before_each(state=None, configuration=None, info={}):
    global env
    steps = [] if state == None else [state]
    env = make("football", steps=steps,
               configuration=configuration, info=info, debug=True)


def test_to_json():
    before_each()
    json = env.toJSON()
    assert json["name"] == "football"
    assert json["rewards"] == [0.0, 0.0]
    assert json["statuses"] == ["ACTIVE", "ACTIVE"]
    assert json["specification"]["reward"]["type"] == ["number", "null"]

def clear_players_raw(state):
  state = copy.deepcopy(state)
  for entry in state:
    entry.observation.players_raw = []
  return state


def test_single_agent():
    before_each(configuration={"team_1": 1, "team_2": 0, "scenario_name": "11_vs_11_stochastic"})
    x = env.reset()

    assert clear_players_raw(env.reset()) == [
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        },
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 0,
                "players_raw": []
            }
        }
    ]

    # Correct step from agent 0.
    assert clear_players_raw(env.step([[0],[]])) == [
        {
            "action": [0],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        },
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 0,
                "players_raw": []
            }
        }
    ]

    # Incorrect step from agent 1 (it is forbidden to act in this scenario,
    # as 'team_2' players is set to 0).
    assert clear_players_raw(env.step([[0],[1]])) == [
        {
            "action": [0],
            "status": "DONE",
            "reward": 100,
            'info': {'debug_info': 'Opponent made invalid move. You win.'},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        },
        {
            "action": [1],
            "status": "INVALID",
            "reward": None,
            'info': {'debug_info': 'Invalid number of actions provided: Expected 0, got 1.'},
            "observation": {
                "controlled_players": 0,
                "players_raw": []
            }
        }
    ]


def test_multi_agent():
    before_each(configuration={"team_1": 2, "team_2": 1, "scenario_name": "11_vs_11_stochastic"})
    x = env.reset()

    assert clear_players_raw(env.reset()) == [
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 2,
                "players_raw": []
            }
        },
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        }
    ]

    # Correct step from both agents.
    assert clear_players_raw(env.step([[0, 2],[4]])) == [
        {
            "action": [0, 2],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 2,
                "players_raw": []
            }
        },
        {
            "action": [4],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        }
    ]

    # Incorrect step from agent 1  - too many actions passed.
    assert clear_players_raw(env.step([[0, 1, 2], [1]])) == [
        {
            "action": [0, 1, 2],
            "status": "INVALID",
            "reward": None,
            'info': {'debug_info': 'Invalid number of actions provided: Expected 2, got 3.'},
            "observation": {
                "controlled_players": 2,
                "players_raw": []
            }
        },
        {
            "action": [1],
            "status": "DONE",
            "reward": 100,
            'info': {'debug_info': 'Opponent made invalid move. You win.'},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        }
    ]

def test_deadline():
    before_each(configuration={"team_1": 1, "team_2": 0, "scenario_name": "11_vs_11_stochastic"})
    x = env.reset()

    assert clear_players_raw(env.reset()) == [
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        },
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 0,
                "players_raw": []
            }
        }
    ]

    # Correct step from agent 0.
    assert clear_players_raw(env.step([DeadlineExceeded(),[]])) == [
        {
            "action": None,
            "status": "TIMEOUT",
            "reward": None,
            "info": {},
            "observation": {
                "controlled_players": 1,
                "players_raw": []
            }
        },
        {
            "action": [],
            "status": "ACTIVE",
            "reward": 0,
            "info": {},
            "observation": {
                "controlled_players": 0,
                "players_raw": []
            }
        }
    ]


def test_render():
    video_file = "/tmp/video.webm"
    before_each(configuration={"team_1": 1, "team_2": 1, "scenario_name": "tests.penalty", "save_video": True},
        info={"LiveVideoPath": video_file})
    env.step([[0],[0]])
    env.render(mode="human", width=800, height=600)
    output = env.step([[0],[0]])
    env.render(mode="human", width=800, height=600)
    while output[0]['status'] == 'ACTIVE':
        output = env.step([[0],[0]])
    assert os.path.isfile(video_file)



import random
from .utils import get_score


def rock(observation, configuration):
    return 0


def paper(observation, configuration):
    return 1


def scissors(observation, configuration):
    return 2


def copy_opponent(observation, configuration):
    if observation.step > 0:
        return observation.lastOpponentAction
    else:
        return random.randrange(0, configuration.signs)


last_react_action = None


def reactionary(observation, configuration):
    global last_react_action
    if observation.step == 0:
        last_react_action = random.randrange(0, configuration.signs)
    elif get_score(last_react_action, observation.lastOpponentAction) <= 1:
        last_react_action = (observation.lastOpponentAction + 1) % configuration.signs

    return last_react_action


last_counter_action = None


def counter_reactionary(observation, configuration):
    global last_counter_action
    if observation.step == 0:
        last_counter_action = random.randrange(0, configuration.signs)
    elif get_score(last_counter_action, observation.lastOpponentAction) == 1:
        last_counter_action = (last_counter_action + 2) % configuration.signs
    else:
        last_counter_action = (observation.lastOpponentAction + 1) % configuration.signs

    return last_counter_action


action_histogram = {}


def statistical(observation, configuration):
    global action_histogram
    if observation.step == 0:
        action_histogram = {}
        return
    action = observation.lastOpponentAction
    if action not in action_histogram:
        action_histogram[action] = 0
    action_histogram[action] += 1
    mode_action = None
    mode_action_count = None
    for k, v in action_histogram.items():
        if mode_action_count is None or v > mode_action_count:
            mode_action = k
            mode_action_count = v
            continue

    return (mode_action + 1) % configuration.signs


agents = {
    "rock": rock,
    "paper": paper,
    "scissors": scissors,
    "copy_opponent": copy_opponent,
    "reactionary": reactionary,
    "counter_reactionary": counter_reactionary,
    "statistical": statistical
}


from kaggle_environments import make
from .halite import random_agent
from .helpers import *


def test_halite_no_repeated_steps():
    step_count = 10
    actual_steps = []

    def step_appender_agent(obs, config):
        actual_steps.append(obs.step)
        return {}

    env = make("halite", configuration={"episodeSteps": step_count}, debug=True)
    env.run([step_appender_agent])
    assert actual_steps == list(range(step_count - 1))


def test_halite_completes():
    env = make("halite", configuration={"episodeSteps": 100})
    env.run([random_agent, random_agent])
    json = env.toJSON()
    assert json["name"] == "halite"
    assert json["statuses"] == ["DONE", "DONE"]


def test_halite_exception_action_has_error_status():
    env = make("halite")

    def error_agent(obs, config):
        raise Exception("An exception occurred!")
    env.run([error_agent, random_agent])
    json = env.toJSON()
    assert json["name"] == "halite"
    assert json["statuses"] == ["ERROR", "DONE"]


def test_halite_helpers():
    env = make("halite", configuration={"size": 3})

    @board_agent
    def helper_agent(board):
        for ship in board.current_player.ships:
            ship.next_action = ShipAction.NORTH
        for shipyard in board.current_player.shipyards:
            shipyard.next_action = ShipyardAction.SPAWN

    env.run([helper_agent, helper_agent])

    json = env.toJSON()
    assert json["name"] == "halite"
    assert json["statuses"] == ["DONE", "DONE"]


def create_board(size=3, starting_halite=0, agent_count=2, random_seed=0):
    env = make("halite", configuration={
        "size": size,
        "startingHalite": starting_halite,
        "randomSeed": random_seed
    })
    return Board(env.reset(agent_count)[0].observation, env.configuration)


def test_move_moves_ship():
    size = 3
    board = create_board(size, agent_count=1)
    for ship in board.current_player.ships:
        ship.next_action = ShipAction.SOUTH
    next_board = board.next()
    for ship in board.ships.values():
        next_position = ship.position.translate(Point(0, -1), size)
        next_ship = next_board.ships[ship.id]
        assert next_ship.position == next_position


def move_toward(ship, target: Point):
    (x1, y1) = ship.position
    (x2, y2) = target
    if x2 > x1:
        return ShipAction.EAST
    elif x2 < x1:
        return ShipAction.WEST
    elif y2 > y1:
        return ShipAction.NORTH
    elif y2 < y1:
        return ShipAction.SOUTH


def test_equal_ship_collision_destroys_both_ships():
    size = 3
    board = create_board(size, agent_count=2)
    for ship in board.ships.values():
        ship.next_action = move_toward(ship, Point(1, 1))
    next_board = board.next()
    assert len(next_board.ships) == 0


def test_unequal_ship_collision_destroys_weaker_ship():
    board = create_board(agent_count=2)
    for opponent in board.opponents:
        for ship in opponent.ships:
            # Make the opponents' ships have more halite so they'll be destroyed
            ship._halite = 1000
    for ship in board.ships.values():
        ship.next_action = move_toward(ship, Point(1, 1))
    next_board = board.next()
    assert len(next_board.current_player.ships) == 1
    assert len(next_board.ships) == 1


def first(iterable):
    return next(iter(iterable))


def test_ship_shipyard_collision_destroys_both():
    board = create_board(agent_count=2)
    player_ship = first(board.current_player.ships)
    opponent_ship = first(first(board.opponents).ships)
    opponent_ship.next_action = ShipAction.CONVERT
    board = board.next()
    assert len(board.ships) == 1
    assert len(board.shipyards) == 1
    while player_ship.id in board.ships:
        board.ships[player_ship.id].next_action = move_toward(player_ship, opponent_ship.position)
        board = board.next()
    assert len(board.ships) == 0
    assert len(board.shipyards) == 0


def test_cells_regen_halite():
    board = create_board(starting_halite=1000, agent_count=1)
    cell = first(board.cells.values())
    next_board = board.next()
    next_cell = next_board[cell.position]
    expected_regen = round(cell.halite * board.configuration.regen_rate, 3)
    # We compare to a floating point value here to handle float rounding errors
    assert next_cell.halite - cell.halite - expected_regen < .000001


def test_no_move_on_halite_gathers_halite():
    board = create_board(starting_halite=1000, agent_count=1)
    ship = first(board.ships.values())
    expected_delta = int(ship.cell.halite * board.configuration.collect_rate)
    next_board = board.next()
    next_ship = next_board.ships[ship.id]
    ship_delta = next_ship.halite - ship.halite
    cell_delta = round(ship.cell.halite - next_ship.cell.halite, 3)
    assert ship_delta == expected_delta
    assert cell_delta == expected_delta


def test_move_on_halite_gathers_no_halite():
    board = create_board(starting_halite=1000, agent_count=1)
    ship = first(board.ships.values())
    ship.next_action = ShipAction.NORTH
    next_board = board.next()
    next_ship = next_board.ships[ship.id]
    ship_delta = next_ship.halite - ship.halite
    assert ship_delta == 0


def test_failed_convert_gathers_halite():
    board = create_board(starting_halite=1000, agent_count=1)
    board.current_player._halite = board.configuration.convert_cost - 1
    ship = first(board.ships.values())
    ship.next_action = ShipAction.CONVERT
    expected_delta = int(ship.cell.halite * board.configuration.collect_rate)
    next_board = board.next()
    next_ship = next_board.ships[ship.id]
    ship_delta = next_ship.halite - ship.halite
    cell_delta = round(ship.cell.halite - next_ship.cell.halite, 3)
    assert ship_delta == expected_delta
    assert cell_delta == expected_delta


def test_shipyard_ids_not_reused():
    board = create_board(starting_halite=1000, agent_count=1)
    ship = first(board.ships.values())
    ship.next_action = ShipAction.CONVERT
    board = board.next()
    shipyard = board.cells[ship.position].shipyard
    assert ship.id != shipyard.id


def test_seed_parameter():
    seed = 9

    def aggregate_halite_for_board(seed):
        board = create_board(starting_halite=1000, agent_count=1, random_seed=seed)
        return sum(map(lambda c: c.halite, board.cells.values()))

    assert aggregate_halite_for_board(seed) == aggregate_halite_for_board(seed)



import json
from os import path
from .agents import agents as all_agents
from .utils import get_score


def interpreter(state, env):
    player1 = state[0]
    player2 = state[1]

    # Specification can fully handle the reset.
    if env.done:
        return state

    step = len(env.steps)
    player1.observation.step = step

    def is_valid_action(player, sign_count):
        return (
            player.action is not None and
            isinstance(player.action, int) and
            0 <= player.action < sign_count
        )

    # Check for validity of actions
    is_player1_valid = is_valid_action(player1, env.configuration.signs)
    is_player2_valid = is_valid_action(player2, env.configuration.signs)
    if not is_player2_valid:
        player2.status = "INVALID"
        player2.reward = 0

        if is_player1_valid:
            player1.status = "DONE"
            player1.reward = 1
            return state

    if not is_player1_valid:
        player1.status = "INVALID"
        player1.reward = 0

        if is_player2_valid:
            player2.status = "DONE"
            player2.reward = 1
            return state
        else:
            return state

    score = get_score(player1.action, player2.action)
    player1.observation.lastOpponentAction = player2.action
    player1.reward += score
    player2.observation.lastOpponentAction = player1.action
    player2.reward += -score
    remaining_steps = env.configuration.episodeSteps - step - 1
    if abs(player1.reward - player2.reward) > remaining_steps * 2:
        player1.status = "DONE"
        player2.status = "DONE"
    elif remaining_steps <= 0:
        player1.status = "DONE"
        player2.status = "DONE"
    return state


def renderer(state, env):
    sign_names = ["Rock", "Paper", "Scissors", "Spock", "Lizard"]
    rounds_played = len(env.steps)
    board = ""

    # This line prints results each round, good for debugging
    for i in range(1, rounds_played):
        step = env.steps[i]
        right_move = step[0].observation.lastOpponentAction
        left_move = step[1].observation.lastOpponentAction
        board += f"Round {i}: {sign_names[left_move]} vs {sign_names[right_move]}, Score: {step[0].reward} to {step[1].reward}\n"

    board += f"Game ended on round {rounds_played - 1}, final score: {state[0].reward} to {state[0].reward}\n"
    return board


dir_path = path.dirname(__file__)
json_path = path.abspath(path.join(dir_path, "rps.json"))
with open(json_path) as json_file:
    specification = json.load(json_file)


def html_renderer():
    js_path = path.abspath(path.join(dir_path, "rps.js"))
    with open(js_path, encoding="utf-8") as js_file:
        return js_file.read()


agents = all_agents